##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


from sensiml.datamanager.pipeline import PipelineStep


class QueryCall(PipelineStep):
    """The base class for a query call."""

    def __init__(self, name):
        super(QueryCall, self).__init__(name=name, step_type='QueryCall')
        self._query = None

    @property
    def query(self):
        return self._query

    @query.setter
    def query(self, query):
        self._query = query

    @property
    def outputs(self):
        return self._outputs

    @outputs.setter
    def outputs(self, outputs):
        self._outputs = outputs

    def _to_dict(self):
        query_dict = {}
        query_dict['name'] = self.query.name
        query_dict['type'] = 'query'
        query_dict['outputs'] = getattr(self, '_outputs')
        return query_dict



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017-18  SensiML Corporation.                                   #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


import os
import json
import IPython
from collections import namedtuple
from operator import attrgetter
from pandas import DataFrame
from ipywidgets import widgets
from ipywidgets import Layout, Button, VBox, HBox, Box, FloatText, Textarea,\
    Dropdown, Label, IntSlider, Checkbox, Text, Button, SelectMultiple, HTML
from IPython.display import display
from ipywidgets import IntText
from json import dumps as jdump
from sensiml.widgets.base_widget import BaseWidget
from sensiml.widgets.renderers import WidgetAttributeRenderer


category_item_layout = Layout(
    # display='flex',
    size=16,
    #border='solid 2px',
    justify_content='flex-start',
    #background_color= 'red',
    overflow='visible'
)


def clean_name(name):
    return ''.join(e if e.isalnum() else '_' for e in name)


class DownloadWidget(BaseWidget):
    def __init__(self, dsk=None, level='Project', folder='knowledgepacks'):
        self._dsk = dsk
        if not os.path.exists(folder):
            os.makedirs(folder)
        self.setup(level=level)

    def setup(self, level):
        self.kb_description = {'parent': {},
                               'sub': {'Report': 'Report'}}

        self.kb_dict = {'parent': [],
                        'sub': []}
        self.level = level

    def select_platform(self, b):
        platform = self._dsk.platforms.get_platform_by_id(
            self._widget_platform.value)
        supported_source_drivers = platform.supported_source_drivers
        supported_source_keys = {x: x for x in list(
            supported_source_drivers.keys())}

        capture_configurations = {
            k: v.uuid for k, v in self._dsk.project.capture_configurations.build_capture_list().items()}

        supported_source_keys.update(capture_configurations)
        supported_source_keys['Custom'] = 'Custom'

        default_source_driver = supported_source_drivers.get("Default", None)
        supported_outputs = platform.supported_outputs

        if default_source_driver is not None:
            # Remove the Default option from the list.
            supported_source_keys.pop("Default")

        self._widget_source.options = supported_source_keys
        if len(self._widget_source.options) > 0:
            if capture_configurations:
                self._widget_source.value = next(
                    iter(capture_configurations.items()))[1]
            else:
                self._widget_source.value = 'Custom'
        else:
            self._widget_source.options = {'Custom': 'Custom'}
            self._widget_source.value = self._widget_source.options[0]
        self._widget_app_outputs.options = supported_outputs
        self._widget_app_outputs.value = supported_outputs
        self._change_rate("value")
        if len(self._widget_source_rate.options) > 0:
            self._widget_source_rate.value = self._widget_source_rate.options[0]
        else:
            self._widget_source_rate.options = ['']
            self._widget_source_rate.value = ''
        self._widget_target_os.options = platform.target_os_options
        self._widget_target_os.value = self._widget_target_os.options[0]

        if platform.can_build_binary:
            self._widget_download_type.options = ["Binary", "Library"]
        else:
            self._widget_download_type.options = ["Library"]
        self._widget_download_type.value = self._widget_download_type.options[0]

    def generate_description(self, b):
        if hasattr(self._dsk, '_auto_sense_ran'):
            if self._dsk._auto_sense_ran:
                self.renderer.render(
                    "KnowledgePack List Needs updating. Updating Now... Reselect and redownload.")
                self._refresh_models_list(None)
                self._dsk._auto_sense_ran = False
                return

        platform = self._dsk.platforms.get_platform_by_id(
            self._widget_platform.value)
        supported_outputs = platform.supported_outputs
        parent_name = None
        parent_uuid = None
        description = {}
        self.set_parent_model(None)
        if not self.kb_description['parent']:
            self.renderer.render("No Parent Model Selected.")
            return
        for key in self.kb_description['parent']:
            description = {clean_name(key): {'uuid': self.kb_description['parent'][key],
                                             'results': {},
                                             'source': self._widget_source.value}}
            parent_name = clean_name(key)
            parent_uuid = self.kb_description['parent'][key]
        for parent in self.kb_dict['parent'][0][1:]:
            description[parent_name]['results'].update(
                {format(parent.description.split('-')[0]): clean_name(parent.value)})
        for key in self.kb_description['sub']:
            if key != 'Report':
                sub_description = {'uuid': self.kb_description['sub'][key],
                                   'parent': parent_name,
                                   'segmenter_from': 'parent', }
                description.update({clean_name(key): sub_description})
        board_name = self._dsk.platforms.get_platform_by_id(
            self._widget_platform.value).board_name

        kp_uuid = parent_uuid
        kp_platform = self._widget_platform.value
        kp_debug = self._widget_debug.value
        kp_test_data = self._widget_test_data.value
        kp_download_type = self._widget_download_type.value
        sample_rate = self._widget_source_rate.value if self._widget_source_rate.value in self._widget_source_rate.options else 100
        output_options = [x.lower() for x in self._widget_app_outputs.value]

        execution_parameters = platform.execution_parameters

        if 'ble' in supported_outputs and 'ble' not in output_options:
            output_options.insert('ble')

        kp_application = execution_parameters.get("application", "Default")

        if kp_uuid is not None:
            kp = self._dsk.get_knowledgepack(kp_uuid)
        else:
            return None

        if kp.knowledgepack_description is not None:
            description = kp.knowledgepack_description
            description['Parent']['source'] = self._widget_source.value

        config = {'target_platform': kp_platform,
                  'test_data': kp_test_data,
                  'debug': kp_debug,
                  'application': kp_application,
                  'sample_rate': sample_rate,
                  'output_options': output_options,
                  'kb_description': description}
        print(config)

        self.renderer.render("Generating Knowledge Pack")

        if kp_download_type == 'Library':
            save_path = kp.download_library(
                config=config, folder='knowledgepacks')
        if kp_download_type == 'Binary':
            save_path = kp.download_binary(
                config=config, folder='knowledgepacks')

        if save_path:
            self.renderer.render(
                "Knowledge Pack saved to {}".format(save_path))

    def set_parent_model(self, b):

        if self._widget_parent_select.value is None:
            return

        kp = self._dsk.get_knowledgepack(self._widget_parent_select.value)
        kp_item = []
        kp_item.append(Label(value=kp.name))

        for key, value in kp.class_map.items():
            kp_item.append(Dropdown(options=self.kb_description['sub'],
                                    description='{} - {}'.format(key, value)))

        self.kb_description['parent'] = {kp.name: kp.uuid}
        self.kb_dict['parent'] = [kp_item]
        self.update_models()

    def get_kp_dict(self):
        if self.level.lower() == 'project':
            kps = self._dsk.project.list_knowledgepacks()
        elif self.level.lower() == 'pipeline':
            kps = self._dsk.pipeline.list_knowledgepacks()
        else:
            kps = self._dsk.list_knowledgepacks()

        if isinstance(kps, DataFrame) and len(kps):
            kps = sorted([(name, value) for name, value in kps[[
                         'Name', 'kp_uuid']].values if name], key=lambda s: s[0].lower())

            return kps

        return [('', None)]

    @staticmethod
    def flatten(l):
        return [item for sublist in l for item in sublist]

    def get_model_list(self):
        return [Label(value='Parent Model')] + self.flatten(self.kb_dict['parent']) + \
            [Label(value='Sub Model')] + self.flatten(self.kb_dict['sub'])

    def get_feature_file_list(self):
        ff = self._dsk.list_featurefiles(silent=True)
        if ff is not None:
            return list(ff['Name'].values)
        else:
            return []

    def get_platform_names(self):
        named_platform = namedtuple('Platform', ['name', 'version', 'id'])
        platform_list = []
        for platform in self._dsk.platforms.platform_list:
            platform_list.append(named_platform(
                platform.platform, platform.platform_version, platform.id))
        platform_list = sorted(
            platform_list, key=attrgetter('name', 'version'))
        ret_dict = {}
        for platform in platform_list:
            ret_dict['{} {}'.format(
                platform.name, platform.version)] = platform.id

        return ret_dict

    def _refresh(self):
        if self._dsk is None:
            return
        self._widget_platform.options = self.get_platform_names()
        #self._widget_platform.value = 10
        self._widget_platform.observe(self.select_platform)
        self.select_platform(None)
        self._widget_test_data.options = [None] + self.get_feature_file_list()
        self._widget_parent_select.options = self.get_kp_dict()
        self._widget_parent_select.value = self._widget_parent_select.options[0][1]
        self._widget_class_map.options = self._get_class_map()

    def _clear(self):
        self._widget_parent_select.options = ['']
        self._widget_parent_select.value = ''
        self._widget_class_map.options = []

    def _update_class_map(self, *args):
        self._widget_class_map.options = self._get_class_map()

    def _get_class_map(self):

        if self._widget_parent_select.value:
            class_map = self._dsk.get_knowledgepack(
                self._widget_parent_select.value).class_map
            return sorted(['{} - {}'.format(key, value) for key, value in class_map.items()])
        else:
            return ''

    def _refresh_models_list(self, b):
        if self._dsk:
            if self._dsk.pipeline:
                self._widget_parent_select.options = self.get_kp_dict()
                self._widget_parent_select.value = self._widget_parent_select.options[0][1]
                self._widget_class_map.options = self._get_class_map()

    def update_models(self):
        if self._dsk is None:
            return

        if self.kb_dict['parent']:
            for output in self.kb_dict['parent'][0][1:]:
                output.options = [
                    k for k, v in self.kb_description['sub'].items()]

    def _change_rate(self, change):
        platform = self._dsk.platforms.get_platform_by_id(
            self._widget_platform.value)
        supported_source_drivers = platform.supported_source_drivers

        if self._widget_source.value == 'Custom':
            self._widget_source_rate.options = ['']
            self._widget_source_rate.value = None
        else:
            self._widget_source_rate.options = supported_source_drivers.get(
                self._widget_source.value, [''])

        if self._widget_source_rate.options[0] == '':
            self._widget_source_rate.layout.visibility = "hidden"
        else:
            self._widget_source_rate.layout.visibility = "visible"

    def create_widget(self):

        self._button_generate = Button(
            icon='download',  tooltip='Generate and Download', layout=Layout(width='15%'))
        self._button_refresh = Button(icon='refresh', layout=Layout(
            width='15%'), tooltip='Refresh Model List')
        self._widget_platform = Dropdown(description="HW Platform", options=[])
        self._widget_target_os = Dropdown(description="Target OS")
        self._widget_app_outputs = SelectMultiple(
            description="Output", options=[])
        self._widget_download_type = Dropdown(description="Format", options=[])
        self._widget_source = Dropdown(
            description="Data Source", options=[])
        self._widget_source_rate = Dropdown(
            description="Sample Rate", options=[])
        self._widget_debug = Dropdown(
            description="Debug", options=[True, False])
        self._widget_test_data = Dropdown(
            description="Test Data", options=[None])
        self._widget_parent_select = Dropdown(
            description="Model Name", options=[], layout=Layout(width='85%'))
        self._widget_class_map = SelectMultiple(
            description="Class Map", options=[''])

        self._widget_render_space = HTML()

        self.renderer = WidgetAttributeRenderer(
            self._widget_render_space, 'value')

        self.kb_items = VBox([
            HBox([self._widget_parent_select, self._button_refresh,


                  self._button_generate,

                  ]),
            self._widget_class_map,
            Label(value='Device Settings', layout=category_item_layout),
            HBox([VBox([self._widget_platform,
                        self._widget_target_os,
                        self._widget_download_type,
                        ]),
                  VBox(
                [self._widget_debug,
                 self._widget_test_data,
                 self._widget_source,
                 self._widget_source_rate,
                 ]),
                self._widget_app_outputs,
            ]),
            self._widget_render_space,
        ])

        self._button_generate.on_click(self.generate_description)
        self._button_refresh.on_click(self._refresh_models_list)
        self._widget_source.observe(self._change_rate, names='value')
        self._widget_parent_select.observe(self._update_class_map)

        self._refresh()

        return self.kb_items



from ipywidgets import Layout, Button, VBox, HBox, Dropdown, Checkbox, Label, widgets, HTML
from sensiml.widgets.base_widget import BaseWidget
import qgrid
from pandas import DataFrame


class PipelineBuilderWidget(BaseWidget):

    def _refresh(self, b=None):
        if not self._dsk:
            raise Exception("Must have dsk ")

        if not self._dsk.project:
            raise Exception("Must have active project")

        if not self._dsk.pipeline:
            raise Exception("Must have active pipeline.")

    def _clear(self):
        for index, subtype in enumerate(self._w_subtype):
            subtype.children[0].value = False

    def _on_button_clicked(self, change=None):
        active_subtypes = get_active_subtypes(self._w_subtypes)
        subtype_list = build_subtype_list(self._dsk, active_subtypes)

        self._widget_render_space.value = generate_pipeline_html_call(
            self._dsk, subtype_list)

    def create_widget(self):

        self._b_run = Button(icon='play', description='Generate Call', layout=Layout(
            width='98%', align_self='flex-end'))

        self._b_run.on_click(self._on_button_clicked)
        self._refresh()

        self._w_subtypes = build_widgets(self._dsk)

        self._widget_render_space = HTML()

        return VBox(self._w_subtypes+[self._b_run, self._widget_render_space])


def get_subtype_defaults(dsk, subtype):
    fgs = dsk.list_functions(functype='Feature Generator',
                             subtype=subtype, kp_functions=True).df.NAME
    params = {}
    for fg in fgs:
        for param in dsk.functions._function_list[fg].input_contract:
            if param['name'] in ['input_data', 'columns', 'group_columns']:
                continue
            params[param['name']] = param['type']

    return params


def build_widgets(dsk):
    w_subtypes = []
    fg_subtypes = dsk.list_functions(
        functype='Feature Generator').df.SUBTYPE.unique()
    for subtype in fg_subtypes:
        subtype_and_params = [widgets.Checkbox(value=False,
                                               description=subtype,
                                               )]
        w_subtypes.append(HBox(subtype_and_params))
    return w_subtypes


def get_active_subtypes(w_subtype):
    active = []
    for subtype in w_subtype:
        if subtype.children[0].value:
            active.append(subtype.children[0].description)
    return active


def build_pipeline(dsk, active_subtypes):
    subtype_list = []
    for subtype in active_subtypes:
        tmp_dict = {'subtype_call': subtype}
        subtype_list.append(tmp_dict)

    dsk.pipeline.add_feature_generator(subtype_list, function_defaults={
                                       'columns': dsk.pipeline.data_columns})


def build_subtype_list(dsk, active_subtypes):
    subtype_list = []
    for subtype in active_subtypes:
        tmp_dict = {'subtype_call': subtype,
                    'params': get_subtype_defaults(dsk, subtype)}
        subtype_list.append(tmp_dict)

    return subtype_list


def generate_pipeline_call(dsk, subtype_list):

    return "dsk.pipeline.add_feature_generator(\n\t\t{0},\n\t\t function_defaults={{'columns':{1}}})".format(',\n\t\t'.join([str(x) for x in subtype_list]), dsk.pipeline.data_columns)


def generate_pipeline_html_call(dsk, subtype_list):

    return "dsk.pipeline.add_feature_generator(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;{0},<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;function_defaults={{'columns':{1}}})".format(',<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'.join([str(x) for x in subtype_list]), dsk.pipeline.data_columns)



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################

from sensiml.datamanager.pipeline import PipelineStep


class TrainAndValidationCall(PipelineStep):
    """The base class for a train-and-validation call."""

    def __init__(self, name=''):
        super(TrainAndValidationCall, self).__init__(
            name=name, step_type='TrainAndValidationCall')
        self._validation_methods = []
        self._classifiers = []
        self._optimizers = []
        self._input_data = ''
        self._feature_table = ''
        self._label_column = None
        self._ignore_columns = None
        self._outputs = []
        self._validation_seed = None

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def validation_methods(self):
        return self._validation_methods

    @validation_methods.setter
    def validation_methods(self, value):
        self._validation_methods = value

    @property
    def classifiers(self):
        return self._classifiers

    @classifiers.setter
    def classifiers(self, value):
        self._classfiers = value

    @property
    def optimizers(self):
        return self._optimizers

    @optimizers.setter
    def optimizers(self, value):
        self._optimizers = value

    @property
    def input_data(self):
        return self._input_data

    @input_data.setter
    def input_data(self, value):
        self._input_data = value

    @property
    def feature_table(self):
        return self._feature_table

    @feature_table.setter
    def feature_table(self, value):
        self._feature_table = value

    @property
    def label_column(self):
        return self._label_column

    @label_column.setter
    def label_column(self, value):
        self._label_column = value

    @property
    def ignore_columns(self):
        return self._ignore_columns

    @ignore_columns.setter
    def ignore_columns(self, value):
        self._ignore_columns = value

    @property
    def outputs(self):
        return self._outputs

    @outputs.setter
    def outputs(self, value):
        self._outputs = value

    @property
    def validation_seed(self):
        return self._validation_seed

    @validation_seed.setter
    def validation_seed(self, value):
        self._validation_seed = value

    def add_validation_method(self, validation_method_call):
        """Adds a validation method call to the train-and-validation call.

            Args:
                validation_method_call (ValidationMethodCall): object to append
        """
        self._validation_methods.append(validation_method_call)

    def remove_validation_method(self, validation_method_call):
        """Removes a validation method call from the train-and-validation call.

            Args:
                validation_method_call (ValidationMethodCall): object to remove
        """
        self._validation_methods = [
            f for f in self._validation_methods if f != validation_method_call]

    def add_classifier(self, classifier_call):
        """Adds a classifier call to the train-and-validation call.

            Args:
                classifier_call (ClassifierCall): object to append
        """
        self._classifiers.append(classifier_call)

    def remove_classifier(self, classifier_call):
        """Removes a classifier call from the train-and-validation call.

            Args:
                classifier_call (ClassifierCall): object to remove
        """
        self._classifiers = [
            f for f in self._classifiers if f != classifier_call]

    def add_optimizer(self, optimizer_call):
        """Adds an optimizer call to the train-and-validation call.

            Args:
                optimizer_call (OptimizerCall): object to append
        """
        self._optimizers.append(optimizer_call)

    def remove_optimizer(self, optimizer_call):
        """Removes an optimizer call from the train-and-validation call.

            Args:
                optimizer_call (OptimizerCall): object to remove
        """
        self._optimizers = [f for f in self._optimizers if f != optimizer_call]

    def _to_dict(self):
        d = {}
        d['name'] = self.name
        d['type'] = 'tvo'
        d['input_data'] = self.input_data
        d['feature_table'] = self.feature_table
        d['label_column'] = self.label_column
        d['ignore_columns'] = self.ignore_columns
        d['outputs'] = self._outputs
        d['validation_seed'] = self._validation_seed

        d['optimizers'] = []
        for item in self._optimizers:
            d['optimizers'].append(item._to_dict())

        d['classifiers'] = []
        for item in self._classifiers:
            d['classifiers'].append(item._to_dict())

        d['validation_methods'] = []
        for item in self._validation_methods:
            d['validation_methods'].append(item._to_dict())

        return d



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


import os
import json
from pandas import DataFrame, Series

from sensiml.base import utility


class PlatformDescription(object):
    """Base class for a PlatformDescription object"""
    _id = 0
    _board_name = ''
    _hardware_accelerators = dict()
    _manufacturer = ''
    _can_build_binary = False
    _platform = ''
    _platform_version = ''
    _description = ''
    _supported_source_drivers = {}
    _supported_outputs = {}
    _ota_capable = False
    _target_os_options = {}
    _execution_parameters = {}

    def __init__(self, connection):
        self._connection = connection

    @property
    def board_name(self):
        return self._board_name

    @property
    def manufacturer(self):
        return self._manufacturer

    @property
    def id(self):
        return self._id

    @property
    def hardware_accelerators(self):
        return self._hardware_accelerators

    @property
    def can_build_binary(self):
        return self._can_build_binary

    @property
    def platform(self):
        return self._platform

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def platform_version(self):
        return self._platform_version

    @property
    def ota_capable(self):
        return self._ota_capable

    @property
    def supported_outputs(self):
        return self._supported_outputs

    @property
    def supported_source_drivers(self):
        return self._supported_source_drivers

    @property
    def target_os_options(self):
        return self._target_os_options

    @property
    def execution_parameters(self):
        return self._execution_parameters

    def get_config(self, test_data='', debug=False):
        """ Generates a default configuation dictionary for downloading a knowledgepack using this platform progrmatically

            Args:
                test_data (str): name of test data file to load onto knowledge pack,
                debug (bool): Build option for knowledge pack. Debug mode will have extra printouts to help with debugging issues.

        """

        if self.supported_source_drivers:
            source = self.supported_source_drivers['Default'][0]
        else:
            source = "Custom"

        if self.supported_outputs:
            outputs = [x.lower() for x in self.supported_outputs]
        else:
            outputs = ['serial']

        config = {'target_platform': self.id,
                  'test_data': test_data,
                  "debug": debug,
                  'application': self.execution_parameters.get('application', 'Default'),
                  'output_options': outputs,
                  'sample_rate': self.supported_source_drivers.get('Default', [None, ''])[1],
                  'source': source
                  }

        return config

    def refresh(self):
        """Calls the REST API and populates the local object properties from the server."""
        url = 'platforms/{0}/'.format(self.id)
        response = self._connection.request('get', url, )
        response_data, err = utility.check_server_response(response)
        if err is False:
            self._board_name = response_data['board_name']
            self._hardware_accelerators = response_data['hardware_accelerators']
            self._platform = response_data['platform']
            self._platform_version = response_data['platform_version']
            self._description = response_data['description']
            self._can_build_binary = response_data['can_build_binary']
            self._ota_capable = response_data['ota_capable']
            self._supported_source_drivers = response_data['supported_source_drivers']
            self._supported_outputs = response_data['supported_outputs']
            self._target_os_options = response_data['target_os_options']
            self._execution_parameters = response_data['execution_parameters']

    def initialize_from_dict(self, input_dictionary):
        """Populates a single transform object from a dictionary of properties from the server.

            Args:
                input_dictionary (dict): containing uuid, type, subtype, name, function_in_file, description,
                input_contract, and subtype
        """
        self._id = input_dictionary['id']
        self._board_name = input_dictionary['board_name']
        self._hardware_accelerators = input_dictionary['hardware_accelerators']
        self._platform = input_dictionary['platform']
        self._platform_version = input_dictionary[u'platform_version']
        self._description = input_dictionary['description']
        self._can_build_binary = input_dictionary['can_build_binary']
        self._ota_capable = input_dictionary.get('ota_capable', False)
        self._supported_source_drivers = input_dictionary.get(
            'supported_source_drivers',
            {"Motion": [200, 100, 50, 25], "Audio": [16000], "Default": ["Motion", 100], "Custom": []})
        self._supported_outputs = input_dictionary.get(
            'supported_outputs', ["BLE", "SERIAL", "LED"])
        self._target_os_options = input_dictionary.get(
            'target_os_options', ["ARM GCC Generic"])
        self._execution_parameters = input_dictionary.get(
            'execution_parameters', {"application": "Default"})

    def __dict__(self):
        ret = {
            'Id': int(self.id),
            'Board': self.board_name,
            'Hardware Accelerated': 'Yes' if len(self.hardware_accelerators.items()) > 0 else 'No',
            'Software Platform': self.platform,
            'Platform Version': self.platform_version,
            'Description': self.description,
        }
        return ret

    def __call__(self):
        pd_dict = self.__dict__()
        pseries = Series(pd_dict, index=pd_dict.keys())
        df = DataFrame()
        df = df.append(pseries, ignore_index=True)
        return df.drop(labels=['Id'], axis=1)



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


import logging
import os
import re
from tempfile import NamedTemporaryFile
from appdirs import user_config_dir
import json
from qgrid import show_grid

import numpy as np
import pandas as pd
from pandas import DataFrame

import sensiml.base.utility as utility
from sensiml.datamanager import Functions, PlatformDescriptions, PipelineSeeds, \
    Projects, Segmenters, CaptureConfigurations
from sensiml.datamanager.knowledgepack import get_knowledgepack, get_knowledgepacks
from sensiml.datasets import DataSets
from sensiml.connection import Connection
from sensiml.base.exceptions import *
from sensiml.pipeline import Pipeline
from sensiml.base.snippets import Snippets, function_help
from sensiml.datamanager.captures import CaptureExistsError

config_dir = user_config_dir(__name__.split('.')[0], False)

logger = logging.getLogger('SensiML')


def print_list(func):
    """ This is a wrapper for printing out lists of objects stored in SensiML Cloud """

    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if len(result.keys()) == 0 and kwargs.get('silent', False) == False:
            print("No {} stored on SensiML Cloud for this project.".format(
                ' '.join(func.__name__.split('_')[1:])))
            return None
        if kwargs.get('get_objects', False) is True:
            return result

        tmp_dict = [{'Name': name, 'UUID': item.uuid, 'Created': item.created_at}
                    for name, item in result.items()]

        return DataFrame(tmp_dict, columns=['Name', 'Created', 'UUID'])

    return wrapper


class SensiML(object):
    """Entrance to SensiML Analytic Suite"""

    # Server Minimum Version For Campatibility
    __minversion__ = '2019.1.0'

    def __init__(self, server="https://sensiml.cloud/",
                 path='connect.cfg', use_jedi=False, **kwargs):

        self._project = None
        self._pipeline = None
        self._knowledgepack_list = None
        auth_url = server+"oauth/"

        self._connection = Connection(
            server=server, auth_url=auth_url, path=path, **kwargs)

        self.validate_client_version(kwargs.get("skip_validate", True))
        self.projects = Projects(self._connection)
        self.datasets = DataSets()
        self.functions = Functions(self._connection)
        self.platforms = PlatformDescriptions(self._connection)
        self.seeds = PipelineSeeds(self._connection)
        self.snippets = Snippets(self.list_functions(
            kp_functions=False, qgrid=False), self.functions.function_list, self.seeds)
        self._feature_files = None

        if use_jedi is False:
            self.setup_jedi_false()

    def setup_jedi_false(self):
        """ This is a temporary bug fix in ipython autocomplete """
        try:
            mgc = get_ipython().magic
            mgc('%config Completer.use_jedi = False')
        except:
            pass

    @property
    def knowledgepack_list(self):
        if self._knowledgepack_list is None:
            self.knowledgepack_list = get_knowledgepacks(self._connection)

        return self._knowledgepack_list

    @knowledgepack_list.setter
    def knowledgepack_list(self, value):
        if value is None:
            self._knowledgepack_list = None
        else:
            self._knowledgepack_list = value.rename(columns={'name': 'Name',
                                                             'project_name': 'Project',
                                                             'sandbox_name': 'Pipeline',
                                                             'uuid': 'kp_uuid',
                                                             'created_at': 'Created'})

    def validate_client_version(self, skip_validate=True):
        """ Perform a Validation check to see if this version of SensiML is up to date with the latest. """

        if skip_validate:
            return

        url = 'version'

        response = self._connection.request('get', url)
        response_data, err = utility.check_server_response(response)

        if response_data["SensiML_Python_Library_Windows"] != self.__minversion__:
            print(
                "SensiML Python Client Library is out of date. Update before connecting to the server.")
            print("In order to update sensiml client use 'pip install sensiml -U'.")
            print("For Instrunctions on updating SensiML Client for the Analytic Studio visit https://sensiml.atlassian.net/wiki/spaces/SS/pages/115179614/Updating+SensiML+Client")
            raise Exception("Python Client is out of date.")

    def logout(self, name=None):
        """Logs out of the current connection."""
        if name is None:
            name = self._connection.server_name

        Connection.logout(name)

    def get_function(self, name):
        """Gets a function method call"""
        return self.functions.function_list[name]

    def function_description(self, name):
        """Gets a description of a pipeline function."""
        print(self.functions.create_function_call(name).__doc__)

    def function_help(self, name):
        """Prints a shortened description of a function. """
        print(function_help(self.functions.function_list[name]))

    def list_functions(self, functype=None, subtype=None, kp_functions=False, qgrid=True):
        """Lists all of the functions available on SensiML Cloud

        Returns:
            Dataframe

        Args:
            functype (str, None): Return only functions with the specified type. ie. "Segmenter"
            subtype (str, None): Return only functions with the specified subtype. ie. "Sensor"
            kp_functions (boolean, True): Return only functions that run on tbe loaded to a device.
            Excludes functions such as feature selection and model training.
        """

        df = DataFrame([{'NAME': f.name, 'TYPE': f.type, "DESCRIPTION": f.description.lstrip('\n').lstrip(' '), "SUBTYPE": f.subtype, "KP FUNCTION": f.has_c_version} for f in self.functions.get_functions()]).\
            sort_values(by=['TYPE', 'SUBTYPE']).reset_index(drop=True)[
            ['NAME', 'TYPE', 'SUBTYPE', 'DESCRIPTION', 'KP FUNCTION']]

        if functype:
            df = df[df['TYPE'] == functype]

        if subtype:
            df = df[df['SUBTYPE'] == subtype]

        if kp_functions:
            df = df[df['KP FUNCTION'] == True][
                ['NAME', 'TYPE', 'SUBTYPE', 'DESCRIPTION']]

        if qgrid:
            return show_grid(df.reset_index(drop=True))
        else:
            return df.reset_index(drop=True)

    def delete_project(self):
        """Deletes a project """
        if self._project is not None:
            self._project.delete()

    @print_list
    def list_projects(self, get_objects=False, silent=False):
        """Lists all of the projects on SensiML Cloud

        Returns:
            DataFrame: projects on SensiML Cloud
        """
        return self.projects.build_project_dict()

    def list_segmenters(self):
        if self._project is None:
            print("project must be set to list segmenters.")
            return None

        segmenters = Segmenters(
            self._connection, self._project).get_segmenters()
        if not len(segmenters):
            print("No segmenters stored on the Cloud.")
            return None
        return DataFrame(segmenters)[['name', 'id', 'custom', 'parent']]

    def list_seeds(self):
        """Lists all of the pipeline seeds on SensiML Cloud

        Returns:
            DataFrame: pipeline seeds on SensiML Cloud
        """
        return DataFrame({
            'Name': [s.name for s in self.seeds],
            'Description': [s.description for s in self.seeds],
        })

    def list_knowledgepacks(self, unnamed=False):
        """Lists all of the knowledgepacks associated with a team

        Returns:
            DataFrame: knowledgepacks on SensiML Cloud
        """
        if self.knowledgepack_list is None:
            print("No Knowledgepacks stored for this team on the cloud.")
            return None

        if unnamed:
            return self.knowledgepack_list[['Name', 'Created', 'Project', 'Pipeline', 'kp_uuid']]

        return self.knowledgepack_list[self.knowledgepack_list['Name'] != ''][['Name', 'Created', 'Project', 'Pipeline', 'kp_uuid']]

    def get_knowledgepack(self, uuid):
        """retrieve knowledgepack by uuid from the server associated with current project

        Args:
            uuid (str): unique identifier for knowledgepack

        Returns:
            TYPE: a knowledgepack object
        """

        return get_knowledgepack(uuid, self._connection)

    @property
    def project(self):
        """The active project"""
        return self._project

    @print_list
    def list_featurefiles(self, get_objects=False, silent=False):
        """List all feature and data files for the active project.

        Args:
            get_objects (bool, False): Also return the featurefile objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")
        return self._project._feature_files.build_featurefile_list()

    @print_list
    def list_datafiles(self, get_objects=False, silent=False):
        """List all feature and data files for the active project.

        Args:
            get_objects (bool, False): Also return the featurefile objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")
        return self._project._feature_files.build_featurefile_list()

    @print_list
    def list_captures(self, get_objects=False):
        """List all captures for the active project

        Args:
            get_objects (bool, False): Also return the capture objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")

        return self._project._captures.build_capture_list()

    @print_list
    def list_capture_configurations(self, get_objects=False):
        """List all captures for the active project

        Args:
            get_objects (bool, False): Also return the capture objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")

        return self._project._capture_configurations.build_capture_list()

    @print_list
    def list_sandboxes(self, get_objects=False):
        """List all sandboxes for the active project.

        Args:
            get_objects (bool, False): Also return the sandbox objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")
        return self._project._sandboxes.build_sandbox_list()

    @print_list
    def list_queries(self, get_objects=False):
        """List all queries for the active project.

        Args:
            get_objects (bool, False): Also return the query objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")
        return self._project._queries.build_query_list()

    @project.setter
    def project(self, name):
        self._project = self.projects.get_or_create_project(name)

    @property
    def pipeline(self):
        """The active pipeline"""
        return self._pipeline

    @pipeline.setter
    def pipeline(self, name):
        if self._project is None:
            raise Exception(
                'Project must be set before a pipeline can be created')

        self._pipeline = Pipeline(self, name=name)

    def create_query(self, name, columns=[], metadata_columns=[], metadata_filter='', segmenter=None, label_column='', combine_labels=None, force=False):
        """Create a query to use as input data in a pipeline.

        Args:
            name (str): Name of the query.
            columns (list, optional): Columns to add to the query result.
            metadata_columns (list, optional): Metadata to add to the query result.
            metadata_filter (str, optional): Filter to apply to the query.
            segmenter (int, optional): Segmenter to filter query by.
            force (bool, False): If True overwrite the query on kb cloud.

        Returns:
            object: Returns a query object that was created.
        """
        query = self.project.queries.get_query_by_name(name)
        new = False
        if query is not None and not force:
            raise QueryExistsException(
                'Query already exists. Set force=True to overwrite.')
        elif query is not None and force:
            query.columns.clear()
            query.metadata_columns.clear()
            query.metadata_filter = ''
            query.label_column = ''
        else:
            query = self.project.queries.new_query()
            query.name = name
            new = True

        for col in columns:
            logger.debug('query_column:' + str(col))
            query.columns.add(col)

        for col in metadata_columns:
            logger.debug('query_metadata_column:' + str(col))
            query.metadata_columns.add(col)

        if metadata_filter:
            logger.debug('query_metadata_filter:' + str(metadata_filter))
            query.metadata_filter = metadata_filter

        if label_column:
            query.label_column = label_column

        if combine_labels:
            query.combine_labels = combine_labels

        query.segmenter = segmenter

        if new:
            query.insert()
        else:
            query.update()

        return query

    def get_query(self, name):
        if self.project is None:
            print("Project must be set first")
            return

        return self.project.queries.get_query_by_name(name)

    def upload_data_file(self, name, path, force=False):
        """Upload data source from local CSV file"""
        logger.debug('set_feature_file:' + name + ':' + path)
        print('Uploading file "{}" to SensiML Cloud.'.format(name))
        if name[-4:] != '.csv':
            name = "{}.csv".format(name)

        feature_file = self._project._feature_files.get_featurefile_by_name(
            name)
        if feature_file is None:
            new = True
            feature_file = self._project.featurefiles.new_featurefile()
        else:
            new = False
            if not force:
                raise FeatureFileExistsException()

        feature_file.filename = name
        feature_file.path = path
        if new:
            feature_file.insert()
        else:
            feature_file.update()

        print('Upload of file "{}"  to SensiMLKB  Cloud completed.'.format(name))

        return self

    def upload_dataframe(self, name, dataframe, force=False):
        """Set data source from a pandas DataFrame."""
        logger.debug('set_data:' + name)

        with NamedTemporaryFile(delete=False) as temp:
            dataframe.to_csv(temp.name, index=False)
            logger.debug('set_dataframe:' + name + ':' + temp.name)
            self.upload_data_file(name, temp.name, force=force)

        os.remove(temp.name)

        return self

    def clear_session_cache(self):
        for _, _, filenames in os.walk(config_dir):
            for filename in filenames:
                if re.match(r'_token.json$', filename):
                    os.unlink(filename)

    def upload_project(self, name, dclprojpath, asynchronous=True):
        if name in self.list_projects()['Name'].values:
            print("Project with this name already exists.")
            return

        self.project = name
        capture_metadata = None
        basedir = os.path.dirname(dclprojpath)
        capture_dir = os.path.join(basedir, 'data')
        segment_map = {}

        with open(dclprojpath, 'r') as f:
            dclproj = json.load(f)

        plugin_config = dclproj.get('ProjectCapturePluginConfig', None)
        if plugin_config is not None:
            self.project.plugin_config = plugin_config
            print('Capture Config Found, Updating...')
            self.project.update()

        for segmenter in dclproj.get('ProjectSegmenters'):

            if segmenter.get('parameters', None):
                params = json.loads(segmenter.get('parameters'))
                call = self.functions.create_function_call(params['name'])
                for k, v in params['inputs'].items():
                    setattr(call, k, v)
            else:
                call = None

            new_segmenter = self.project.add_segmenter(segmenter.get('name'),
                                                       call,
                                                       preprocess=segmenter.get(
                'preprocess'),
                custom=segmenter.get('custom'))

            segment_map[segmenter['name']] = new_segmenter['id']
            segment_map[segmenter['id']] = new_segmenter['id']

        capture_files = os.listdir(capture_dir)
        for capture in capture_files:
            print("Uploading Capture data, metadata and lables for {}".format(capture))
            try:

                capture_metadata = {}
                file_metadata = []
                if os.path.exists(os.path.join(basedir, 'metadata', os.path.basename(capture)[:-3]+'dcl')):
                    with open(os.path.join(basedir, 'metadata', os.path.basename(capture)[:-3]+'dcl'), 'r') as f:
                        capture_metadata = json.load(f)
                        file_metadata = capture_metadata.pop('FileMetadata')

                capture_obj = self.project.captures.create_capture(os.path.basename(capture),
                                                                   os.path.join(
                    capture_dir, capture),
                    capture_info=capture_metadata,
                    asynchronous=asynchronous,
                )
                for metadata in file_metadata:
                    print('metadata', metadata['name'], metadata['value'])
                    metadata_obj = capture_obj.metadataset.new_metadata()
                    metadata_obj.value = metadata['value']
                    metadata_obj.name = metadata['name']
                    metadata_obj.insert()

                for segment_dir in os.listdir(os.path.join(basedir, 'segments')):
                    if not os.path.exists(os.path.join(basedir, 'segments',
                                                       segment_dir, os.path.basename(capture)[:-3]+'sdcl')):
                        continue

                    with open(os.path.join(basedir, 'segments', segment_dir, os.path.basename(capture)[:-3]+'sdcl'), 'r') as f:
                        segment_labels = json.load(f).pop('DetectedSegments')

                    for segment in segment_labels:
                        for label in segment['LabelData']:
                            print('label', label['name'], label['value'])
                            label_obj = capture_obj.labelset.new_label()
                            label_obj.sample_start = label['capture_sample_sequence_start']
                            label_obj.sample_end = label['capture_sample_sequence_end']
                            label_obj.value = label['value']
                            label_obj.name = label['name']
                            label_obj.segmenter = segment_map[label['segmenter']]
                            label_obj.insert()

            except CaptureExistsError:
                print("{} already uploaded to this project.".format(capture))



import json
from sensiml.datamanager.capture_configuration import CaptureConfiguration
import sensiml.base.utility as utility


class CaptureConfigurationExistsError(Exception):
    """Base class for an Capture exists error"""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class CaptureConfigurations():
    """Base class for a collection of Captures."""

    def __init__(self, connection, project):
        self._connection = connection
        self._project = project

    def create_capture_configuration(self, name, configuration, uuid=None):
        """Creates an capture object from the given filename and filepath.

            Args:
                name (str): desired name of the file on the server
                configuration (dict): configuration file
                uuid (str): Specify the uuid this configuration will be created with

            Returns:
                capture_configuration object

            Raises:
                CaptureClonfigurationExistsError, if the Capture already exists on the server
        """
        data = {
            'connection': self._connection,
            'project': self._project,
            'name': name,
        }

        if self.get_capture_configuration_by_name(name) is not None:
            raise CaptureConfigurationExistsError(
                "capture configuration {0} already exists.".format(name))
        else:
            instance = CaptureConfiguration.initialize_from_dict(data)
            instance.configuration = configuration
            instance.uuid = uuid
            instance.insert()

        return instance

    def build_capture_list(self):
        """Populates the function_list property from the server."""
        capture_list = {}

        response = self.get_capture_configurations()
        for capture_configuration in response:
            capture_list[capture_configuration.name] = capture_configuration

        return capture_list

    def get_capture_configuration_by_name(self, name):
        """Gets an capture configuration object from the server using its name property.

            Args:
                name (str): the capture configuration's name

            Returns:
                capture configuration  object or None if it does not exist
        """
        capture_list = self.get_capture_configurations()
        for capture in capture_list:
            if capture.name == name:
                return capture
        return None

    def __getitem__(self, key):
        if type(key) == str:
            return self.get_capture_configuration_by_filename(key)
        else:
            return self.get_capture_configurations()[key]

    def _new_capture_configuration_from_dict(self, data):
        """Creates a new capture using the dictionary.

            Args:
                data (dict): dictionary of capture properties

            Returns:
                capture configuration object
        """
        data.update({
            'connection': self._connection,
            'project': self._project
        })
        return CaptureConfiguration.initialize_from_dict(data)

    def get_capture_configurations(self):
        """Gets all capture configurations from the server.

            Returns:
                list of capture configurations for the project
        """
        # Query the server and get the json
        url = 'project/{0}/captureconfiguration/'.format(self._project.uuid)
        response = self._connection.request('get', url,)
        response_data, err = utility.check_server_response(response)

        # Populate each capture from the server
        data = []
        if err is False:
            for params in response_data:
                data.append(self._new_capture_configuration_from_dict(params))

        return data



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


import json
from pandas import DataFrame, Series
from numpy import NaN
import logging
from sensiml.datamanager.confusion_matrix import ConfusionMatrix, ConfusionMatrixException
import sensiml.base.utility as utility

logger = logging.getLogger('ModelMetrics')


class ModelException(Exception):
    pass


def get_average(value):
    if isinstance(value, dict):
        return value['average']
    else:
        return value


class ModelMetrics(object):
    """Base class for a model metrics object.

    Attributes:
        confusion_matrix_stats (list[ConfusionMatrix]): comprehensive metrics returned for the model
        train_set (list): indices of the input data that the model was trained with
        validation_set (list): indices of the input data that the model was validated with
        test_set (list): indices of the input data that the model was tested with
        debug (dict): structure containing debug information for some models
        neurons (list[dict]): model neuron array
        parameters (dict): model parameters
        knowledgepack (KnowledgePack): knowledgepack associated with the model
    """

    def _order_columns(self, confusion_matrix):
        """Orders the columns such that the classes come first, then UNK and UNC."""
        columns = confusion_matrix.index.tolist() + ['UNK', 'UNC']
        return confusion_matrix[columns]

    def __init__(self, configuration, sandbox, index, model_result):
        """Initializes a model result set instance

        Args:
            configuration (dict)
            sandbox (Sandbox)
            index (str)
            model_result (dict)
        """
        self._configuration = configuration
        self._sandbox = sandbox
        self._kp_uuid = model_result['KnowledgePackID'] if 'KnowledgePackID' in model_result else None
        self._index = index
        self._number_of_neurons = model_result['metrics']['validation'].get(
            'NeuronsUsed', 0)
        self._confusion_matrix = {'train': None,
                                  'validation': None, 'test': None}
        self.confusion_matrix_stats = {
            'train': None, 'validation': None, 'test': None}
        self._accuracy = {'train': None, 'validation': None, 'test': None}
        self._f1score = {'train': None, 'validation': None, 'test': None}
        self._sensitivity = {'train': None, 'validation': None, 'test': None}

        for key in ['train', 'test', 'validation']:
            if model_result['metrics'][key]:
                try:
                    self.confusion_matrix_stats[key] = ConfusionMatrix(
                        model_result['metrics'][key])
                except ConfusionMatrixException:
                    pass
                self._confusion_matrix[key] = self._order_columns(DataFrame.from_dict(
                    model_result['metrics'][key]['ConfusionMatrix']).transpose())

                self._accuracy[key] = get_average(
                    model_result['metrics'][key]['accuracy'])
                self._f1score[key] = get_average(
                    model_result['metrics'][key]['f1_score'])
                self._sensitivity[key] = get_average(
                    model_result['metrics'][key]['sensitivity'])

        self.train_set = model_result['train_set']
        self.validation_set = model_result['validation_set']
        self.test_set = model_result['test_set']
        self.debug = model_result['debug']
        self.neurons = model_result['parameters']
        self.parameters = model_result['parameters']
        self._knowledgepack = None
        self._model_result = model_result

    def _get_knowledgepack(self):
        if self._knowledgepack:
            return self._knowledgepack
        elif self._kp_uuid:
            self._knowledgepack = self._sandbox.knowledgepack(self._kp_uuid)
            return self._knowledgepack
        else:
            raise ModelException('This model does not have a knowledgepack.')

    @property
    def knowledgepack(self):
        """The model's KnowledgePack object"""
        return self._get_knowledgepack()

    def summarize(self, metrics_set='validation'):
        """Prints a formatted summary of the model metrics

        Args:
            metrics_set (optional[str]): options are 'validation' (the default), 'test', and 'train'
        """
        if not self._accuracy.get(metrics_set, None):
            print("No results for metric set ", metrics_set)
            return

        formatted_output = ('MODEL INDEX: {}\n'
                            '\tf1-score\ttrain: {:.2f}  validate: {:.2f}\n'
                            '\taccuracy\ttrain: {:.2f}  validate: {:.2f}\n'
                            '\tsensitivity\ttrain: {:.2f}  validate: {:.2f}\n'
                            ).format(self._index, self._accuracy['train'], self._accuracy['validation'],
                                     self._f1score['train'], self._f1score['validation'],
                                     self._sensitivity['train'], self._sensitivity['validation'])

        print(formatted_output)

    def recognize_vectors(self, vectors):
        """Sends a DataFrame of feature vectors to the model's KnowledgePack for recognition.

        Args:
            vectors (DataFrame): where each row is a feature vector with column headings named the same as
            the features generated by the pipeline (order does not matter, but names do) (optional) metadata
            and label columns may be included

        Returns:
             (DataFrame): contains the results of recognition, including predicted class, neuron ID, and distance
        """
        results = []
        kp = self._get_knowledgepack()
        assert len(
            kp.feature_summary) > 0, "Error: No feature summary was found for this Knowledge Pack"

        feature_list = [feature['Feature'] for feature in kp.feature_summary]
        for i, vector in vectors[feature_list].astype(int).iterrows():
            vector_index = int(i)
            vector_dict = {'Vector': list(
                [int(x) for x in vector.values]), 'DesiredResponses': self._number_of_neurons}
            vector_result = kp.recognize_features(vector_dict)
            result = {'Index': vector_index,
                      'Predicted': vector_result['MappedCategoryVector'],
                      'NeuronID': vector_result['NIDVector'][:len(vector_result['MappedCategoryVector'])],
                      'Distance': vector_result['DistanceVector'][:len(vector_result['MappedCategoryVector'])]}
            for column in vectors.columns.difference(feature_list):
                result[column] = vectors.ix[i, column]
            results.append(result)
        df_result = DataFrame(results).set_index('Index', drop=True)
        df_result.index.name = None

        ordered_columns = ['Distance', 'NeuronID', 'Predicted'] + \
                          [i for i in df_result.columns if i not in [
                              'Distance', 'NeuronID', 'Predicted']]
        df_result = df_result[ordered_columns]

        return df_result

    def recognize_signal(self, capturefile=None, stop_step=False,
                         datafile=None, segmenter=True, lock=True, silent=True,
                         platform='emulator', compare_labels=False, **kwargs):
        """Sends a DataFrame of raw signals to be run through the feature generation pipeline and recognized.

        Args:
            capturefile (str): The name of a captured file uploaded throught the data capture lab
            datafile (str): The name of an uploading datafile
            platform (str): "emulator" or "cloud". The "emulator" will run compiled c code giving device exact results,
                 the cloud runs similary to training providing more flexibility in returning early results by setting the
                 stop step.
            stop_step (int): for debugging, if you want to stop the pipeline at a particular step, set stop_step
              to its index
            compare_labels (bool): If there are labels for the input dataframe, use them to create a confusion matrix
            segmenter (bool or FunctionCall): to suppress or override the segmentation algorithm in the original
              pipeline, set this to False or a function call of type 'segmenter' (defaults to True)
            lock (bool, True): If True, waits for the result to return before releasing the ipython cell.


        Returns:
            (DataFrame, dict): a dataframe containing the results of recognition and a dictionary containing the
            execution summary and the confusion_matrix when labels are provided.

            - execution_summary: a summary of which steps ran in the execution engine
            - confusion_matrix: a confusion matrix, only if the input data has a 'Label' column

        """

        kp = self._get_knowledgepack()

        return kp.recognize_signal(capture=capturefile, datafile=datafile,
                                   stop_step=stop_step, segmenter=segmenter, platform=platform, get_result=True, compare_labels=compare_labels)

    def kill_pipeline(self):
        kp = self._get_knowledgepack()
        kp.stop_recognize_signal()

    def __str__(self, metrics_set='validation'):
        output = ('MODEL INDEX: {0}\n'
                  'ACCURACY: {1:.1f}\n'
                  'NEURONS: {2}\n'
                  ).format(self._index, self._accuracy[metrics_set], self._number_of_neurons)
        output += str(self.confusion_matrix_stats[metrics_set])
        return output

    def get_positive_predictive_indexes(self, class_name):
        """
        Returns the indicides of positive predictive classes in the validation set.


        Args:
            class_name (string): name of class to get indices for

        """
        return [self.validation_set[x] for x in self.knowledgepack.get_positive_predictive_indexes(class_name)]

    def get_false_positive_indexes(self, class_name):
        """
        Returns the indicides of false positives in the validation set.

        Args:
            class_name (string): name of class to get indices for


        """
        return [self.validation_set[x] for x in self.knowledgepack.get_false_positive_indexes(class_name)]

    def get_true_positive_indexes(self, class_name):
        """
         Returns the indicides of true positive classification in the validation set.

        Args:
            class_name (string): name of class to get indices for
        """

        return [self.validation_set[x] for x in self.knowledgepack.get_true_positive_indexes(class_name)]


class Configuration(object):

    def __init__(self, result_set, sandbox, config_result):
        """Initializes a configuration instance - the container for all models generated by a TVO configuration."""
        self._result_set = result_set
        self._sandbox = sandbox
        self._training_algorithm = config_result['config']['optimizer']
        self._validation_method = config_result['config']['validation_method']
        self._classifier = config_result['config']['classifier']
        self._metrics = config_result['metrics'][0]
        self._parameters = {k: v for (k, v) in config_result['config'].items() if
                            k not in ['optimizer', 'validation_method', 'classifier']}
        self._models = []
        for index, model in config_result['models'].items():
            self._models.append(ModelMetrics(
                self, self._sandbox, index, model))
        self.sort_models()

    @property
    def models(self):
        return self._models

    @property
    def average_accuracy(self, metrics_set='validation'):
        """Returns the average accuracy for a set of model metrics.

        Args:
            metrics_set (optional[str]): options are 'validation' (the default), 'test', and 'train'
        """
        accuracies = [m._accuracy[metrics_set] for m in self._models]
        return sum(accuracies) / float(len(accuracies))

    def sort_models(self, metric='accuracy', metrics_set='validation'):
        """Sorts the configuration's models by a given metric or property.

        Args:
            metric (optional[str]): options are 'accuracy' (the default), 'index', and 'number_of_neurons'
            metrics_set (optional[str]): options are 'validation' (the default), 'test', and 'train'
        """
        if metric == 'accuracy':
            self._models.sort(
                key=lambda x: x._accuracy[metrics_set], reverse=True)
        elif metric == 'number_of_neurons':
            self._models.sort(key=lambda x: x._number_of_neurons)
        elif metric == 'index':
            self._models.sort(key=lambda x: x._index)

    def get_model_by_index(self, index):
        """Returns the model of the given index (name)."""
        return [model for model in self._models if model._index == index].pop()

    def print_models(self):
        """Prints a summary and confusion matrix of each model in the configuration."""
        for model in self._models:
            print(model)

    def summarize(self, metrics_set='validation', report_number=5):
        """Prints a basic summary of the configuration and its models.

        Args:
            metrics_set (optional[str]): options are 'validation' (the default), 'test', and 'train'
            report_number (int):
        """
        print(self)
        self.metrics()
        print("--------------------------------------\n")
        print("{} MODEL RESULTS : SET {}\n".format(
            self._validation_method.upper(), metrics_set.upper()))
        for index, model in enumerate(self._models):
            if index >= report_number:
                break
            model.summarize(metrics_set)

    def metrics(self):
        average_metrics = ['f1_score', 'precision', 'sensitivity',
                           'f1_score_std', 'precision_std', 'sensitivity_std']

        for metric in average_metrics:
            if self._metrics[metric] is None:
                self._metrics[metric] = 0

        if set(self._metrics.keys()).issuperset(set(average_metrics)):
            print("AVERAGE METRICS:")
            print("F1-SCORE:    {:.1f}   sigma {:.2f}\nSENSITIVITY: {:.1f}   sigma {:.2f}\nPRECISION:   {:.1f}   sigma {:.2f}\n".format(self._metrics['f1_score'], self._metrics['f1_score_std'],
                                                                                                                                        self._metrics['precision'], self._metrics[
                                                                                                                                            'precision_std'],
                                                                                                                                        self._metrics['sensitivity'], self._metrics['sensitivity_std'],))

    def __str__(self):
        return ('TRAINING ALGORITHM: {0}\nVALIDATION METHOD:  {1}\nCLASSIFIER:         {2}\n'
                ).format(self._training_algorithm, self._validation_method, self._classifier)


class ModelResultSet(object):
    """Base class for a model result set object."""

    def __init__(self, project, sandbox):
        """Initializes a model result set instance."""
        self._connection = sandbox._connection
        self._project = project
        self._sandbox = sandbox
        self._results_dict = {}
        self._input_data = None
        self._configurations = []

    @property
    def feature_vectors(self):
        return self._input_data

    @property
    def input_data(self):
        return self._input_data

    @property
    def configurations(self):
        return self._configurations

    def _to_dict(self):
        return self._results_dict

    def _format_configuration(self, config):
        return ('Training Algorithm: {optimizer}\n'
                'Validation Method: {validation_method}\n'
                ).format(**config)

    def sort_models(self, metric='accuracy', metrics_set='validation'):
        """Sorts the models within all configurations by a given metric or property.

        Args:
            metric (optional[str]): options are 'accuracy' (the default), 'index', and 'number_of_neurons'
            metrics_set (optional[str]): options are 'validation' (the default), 'test', and 'train'
        """
        for configuration in self._configurations:
            configuration.sort_models(metric, metrics_set)

    def summarize(self, metrics_set='validation', report_number=5):
        """Prints a basic summary of each configuration and its models.

        Args:
            metrics_set (optional[str]): options are 'validation' (the default), 'test', and 'train'
        """
        for config in self._configurations:
            config.summarize(metrics_set, report_number=report_number)

    def get_knowledgepack_by_index(self, config_index, model_index):
        """Returns the KnowledgePack of the given configuration index and model index."""
        knowledgepack = [kp for kp in self._sandbox.knowledgepack() if
                         kp.configuration_index == config_index and kp.model_index == model_index]
        if knowledgepack:
            return knowledgepack.pop()
        else:
            return 'No knowledgepack found for configuration {0}, model {1}.'.format(config_index, model_index)

    def initialize_from_dict(self, init_dict):
        self._results_dict = init_dict
        if init_dict.get('input_data', None):
            self._input_data = DataFrame.from_dict(
                json.loads(init_dict['input_data']))
        for _, config in self._results_dict['configurations'].items():
            self.configurations.append(
                Configuration(self, self._sandbox, config))



import os
import sys
import io
import json
import IPython
from pandas import DataFrame, read_csv
from ipywidgets import widgets
from ipywidgets import Layout, Tab, BoundedIntText, Button, VBox, HBox, Box, FloatText, Textarea, Dropdown, Label, IntSlider, Checkbox, Text, Image, Button, SelectMultiple, HTML
from IPython.display import display
from ipywidgets import IntText
from json import dumps as jdump
from sensiml.widgets.base_widget import BaseWidget, button_handling
from sensiml.widgets.renderers import WidgetAttributeRenderer
import numpy as np
import matplotlib.pyplot as plt
import qgrid

category_item_layout = Layout(
    # display='flex',
    size=16,
    # border='solid 2px',
    justify_content='flex-start',
    # background_color= 'red',
    overflow='visible'
)


def clean_name(name):
    return ''.join(e if e.isalnum() else '_' for e in name)


class ModelExploreWidget(BaseWidget):
    def __init__(self, dsk=None, level='Project', folder='knowledgepacks'):
        self._dsk = dsk
        self.kp = None
        if not os.path.exists(folder):
            os.makedirs(folder)
        self.setup(level=level)
        self._last_capture_file = None
        self._last_parent_value = None

    def _enable_buttons(self):

        self._button_recognize_signal_capturefile.disabled = False

    def _disable_buttons(self):

        self._button_recognize_signal_capturefile.disabled = True

    def setup(self, level):

        self.level = level

    def get_kp_dict(self):
        if self.level.lower() == 'project':
            kps = self._dsk.project.list_knowledgepacks()
        elif self.level.lower() == 'pipeline':
            kps = self._dsk.pipeline.list_knowledgepacks()
        else:
            kps = self._dsk.list_knowledgepacks()

        if isinstance(kps, DataFrame) and len(kps):
            kps = sorted([(name, value) for name, value in kps[[
                         'Name', 'kp_uuid']].values if name], key=lambda s: s[0].lower())

            return kps

        return [('', None)]

    def _load_results(self, uuid, file_name):

        model_results_folder = 'model_results'
        kp_folder = os.path.join(model_results_folder, uuid)
        results_file_path = os.path.join(kp_folder, file_name)

        if not os.path.exists(model_results_folder):
            os.mkdir(model_results_folder)
            return None

        if not os.path.exists(kp_folder):
            os.mkdir(kp_folder)
            return None

        if not os.path.exists(results_file_path):
            return None

        return read_csv(results_file_path)

    def _cache_results(self, uuid, file_name, results):

        model_results_folder = 'model_results'
        kp_folder = os.path.join(model_results_folder, uuid)
        results_file_path = os.path.join(kp_folder, file_name)

        if not os.path.exists(model_results_folder):
            os.mkdir(model_results_folder)

        if not os.path.exists(kp_folder):
            os.mkdir(kp_folder)

        return results.to_csv(results_file_path)

    def _recognize_signal_all(self, b, recognize=True):

        for capture in self._widget_capture_data.get_selected_df()['capture_name'].values:
            self._recognize_signal(capture)

    @button_handling
    def _recognize_signal(self, capture_file, recognize=True):

        self._last_capture_file = capture_file

        self._output_recognition_classification.layout.visibility = "hidden"

        if not capture_file or not self.kp.uuid:
            return

        if self._widget_parent_select.value != self.kp.uuid:
            return

        self.results = self._load_results(self.kp.uuid, capture_file)

        if self.results is None and not recognize:
            return None

        if self.results is None:
            self._button_recognize_signal_capturefile.description = 'Classifying Signal...'
            self.results, s = self.kp.recognize_signal(
                capture=capture_file, renderer=self.renderer)
            self._button_recognize_signal_capturefile.description = 'RUN'

            if not isinstance(self.results, DataFrame):
                return None

            self._cache_results(self.kp.uuid, capture_file, self.results)

        ytick_labels = ["Unknown"]+[self.kp.class_map[x]
                                    for x in sorted(self.kp.class_map)]

        self.results = post_process(
            self.results, self._w_post_process_steps.value)

        if self.results.shape[0] == 0:
            err = "Post Processing buffer is larger than the number of classifications from capture {}".format(
                capture_file)
            self._widget_render_space.value = err
            print(err)
            return

        ax = self.results.plot(y='Classification', x='SegmentStart', style='--o',
                               figsize=(16, 5), yticks=range(len(ytick_labels)),
                               title='Model: ' + self.kp.name +
                                     '  Capture: '+capture_file +
                                     "  Smoothing: {}".format(
                                         self._w_post_process_steps.value),
                               )

        _ = ax.set_yticklabels(ytick_labels)
        _ = ax.set_xlabel('time (samples)')

        fig = ax.get_figure()

        buf = io.BytesIO()
        fig.savefig(buf, format='png')
        buf.seek(0)

        self._output_recognition_classification.value = buf.read()
        self._output_recognition_classification.layout.visibility = "visible"

        plt.show()

        """
        M = []
        for i in range(self.results.shape[0]):
            M.append(eval(self.results.loc[i]['FeatureVector']))
        columns = []
        for feature in self.kp.feature_summary:
            columns.append(feature['Feature'])
        fv = DataFrame(M, columns=columns)
        fv['Labels'] = self.results.ClassificationName

        # self._dsk.pipeline.visualize_features(
        #    fv, label_column="Labels")

        # self._output_recognition_features.value = buff.read()
        # self._output_recognition_features.visibility = "visible"
        """

    def _rehydrate_model(self, b):

        if self.kp:
            pipeline_steps = self._dsk.pipeline.rehydrate_knowledgepack(
                self.kp, replace=False)

            pipeline_steps.append("r,s = dsk.pipeline.execute()")

            print('\n\n'.join(pipeline_steps))

            instructions = '#### This is a rehydrated Knowledge Pack. \n\n#### Important: Remove the quotes at the start and execute this cell to run. \n\n\n'

            df = DataFrame(
                ['"\n'+instructions+'\n\n'.join(pipeline_steps).replace('"', "\'")+'\n\n\n########""""""'])

            df.to_clipboard(index=False, header=False, excel=True)

            return pipeline_steps

        return ''

    def get_feature_file_list(self):
        ff = self._dsk.list_featurefiles(silent=True)
        if ff is not None:
            return list(ff['Name'].values)
        else:
            return []

    def get_capture_file_list(self):
        ff = self._dsk.list_captures()
        if ff is not None:
            return sorted(list(ff['Name'].values))
        else:
            return []

    def _refresh(self):
        if self._dsk is None:
            return
        project_stats, _ = self._dsk.project.statistics()
        self._widget_capture_data.df = project_stats

        self._widget_parent_select.options = self.get_kp_dict()
        self._widget_parent_select.value = self._widget_parent_select.options[0][1]

    def _clear(self):
        self._widget_parent_select.options = ['']
        self._widget_parent_select.value = ''
        self._widget_confusion_matrix.value = ''
        self._widget_feature_summary.value = ''
        self._widget_model_summary.value = ''
        self._output_recognition_classification.layout.visibility = "hidden"
        self._widget_capture_data.df = DataFrame({'capture_name': ['']})

    def _update_confusion_matrix(self, *args):

        if self._last_parent_value == self._widget_parent_select.value:
            return

        if self._widget_parent_select.value:
            self._last_parent_value = self._widget_parent_select.value
            self.kp = self._dsk.get_knowledgepack(
                self._widget_parent_select.value)
            self._widget_confusion_matrix.value = self.kp.confusion_matrix.__html__()
            self._widget_feature_summary.value = self._get_feature_summary_html(
                self.kp)
            self._widget_model_summary.value = self._get_model_summary_html(
                self.kp)

    def _get_feature_summary_html(self, kp):
        return DataFrame(kp.feature_summary)[
            ['Category', 'Generator', 'Sensors']].to_html()

    def _get_model_summary_html(self, kp):
        model_summary = kp.device_configuration
        model_summary.pop('sample_rate', None)
        model_summary['uuid'] = kp.uuid

        columns = list(model_summary.keys())

        if kp.device_configuration['classifier'] == 'PME':
            important_metrics = ['classifier', 'neurons',
                                 'classification_mode', 'distance_mode', 'optimizer']
            model_summary['optimizer'] = kp.pipeline_summary[-1]['optimizers'][0]['name']
            model_summary['neurons'] = len(kp.neuron_array)
            model_summary['distance_mode'] = 'LSUP' if model_summary['distance_mode'] else 'L1'
            model_summary['classification_mode'] = 'KNN' if model_summary['classification_mode'] else 'RBF'
            columns = [x for x in model_summary.keys() if x not in
                       important_metrics]
            columns = important_metrics+columns

        if kp.device_configuration['classifier'] == 'Decision Tree Ensemble':
            model_summary['n_estimators'] = kp.pipeline_summary[-1]['optimizers'][0]['inputs']['n_estimators']
            model_summary['max_depth'] = kp.pipeline_summary[-1]['optimizers'][0]['inputs']['max_depth']
            columns = ['classifier', 'n_estimators', 'max_depth', 'uuid']

        if kp.device_configuration['classifier'] == 'Boosted Tree Ensemble':
            # TODO: Parse boosted tree ensemble
            columns = list(model_summary.keys())
            pass

        model_summary_df = DataFrame([model_summary])

        return model_summary_df[columns].to_html()

    def _refresh_models_list(self, b):
        if self._dsk:
            if self._dsk.pipeline:
                self._widget_parent_select.options = self.get_kp_dict()
                self._widget_parent_select.value = self._widget_parent_select.options[0][1]

    def create_widget(self):

        description_style = {'description_width': '112px'}

        self._info_validation_method_size = widgets.Button(
            icon="question",
            disabled=True,
            tooltip='Uses a majority vote algorithm over past events in the buffer to determine the current classification. A buffer size of 1 is equivalent to no post processing.',
            layout=Layout(width="5%", align_self='flex-end')
        )

        self._info_run_botton = widgets.Button(
            icon="question",
            disabled=True,
            tooltip='Use the selected model to classify the signals from the captured data files selected above.',
            layout=Layout(width="5%", align_self='flex-end')
        )

        self._button_recognize_signal_capturefile = Button(
            icon='angle-double-right',  tooltip='Run recognition against the provided test data.', layout=Layout(width='30%'), description='Classify')
        self._button_refresh = Button(icon='refresh', layout=Layout(
            width='15%'), tooltip='Refresh Model List')
        self._button_rehydrate = Button(icon='copy', layout=Layout(
            width='15%'), tooltip='Rehydate and paste pipeline coded. Note ( The copied code will have extra """ wrapping the code, remove those before executing')
        self._widget_parent_select = Dropdown(
            description="Model Name", options=[], layout=Layout(width='85%'))
        self._widget_test_data = Dropdown(
            description="Test Data", options=[None])
        self._widget_confusion_matrix = HTML(
            disabled=True,
        )
        self._widget_feature_summary = HTML(
            disabled=True
        )
        self._widget_model_summary = HTML(
            disabled=True
        )
        self._widget_render_space = HTML(
            disabled=True
        )

        self._html_capture_explanation = HTML(
            "Evaluate your model by selecting one or more capture files for classification.")

        self._widget_capture_data = qgrid.show_grid(DataFrame({'capture_name': ['']}),
                                                    grid_options={'rowHeight': 32,
                                                                  'maxVisibleRows': 5,
                                                                  'minVisibleRows': 2,
                                                                  'editable': False,
                                                                  'defaultColumnWidth': 100,
                                                                  "forceFitColumns": True,
                                                                  'sortable': True,
                                                                  'filterable': True},
                                                    column_options={
                                                        'minWidth': 30},
                                                    column_definitions={
                                                        'index': {'width': 30},
                                                        'capture_name': {'width': 350}},
                                                    )

        self.renderer = WidgetAttributeRenderer(
            self._widget_render_space, 'value')

        self._output_recognition_classification = widgets.Image(
            format='png', disabled=True)

        # self._widget_capture_data = Dropdown(
        #    description="Test Data", options=[None],  layout=Layout(width='45%'))

        self._w_post_process_steps = BoundedIntText(
            description='Post Processing', min=1, max=100, value=1, step=5, layout=Layout(width='30%'),
            style={'description_width': '112px'})

        self._button_recognize_signal_capturefile.style.button_color = '#90EE90'

        model_explore_tabs = Tab([
            VBox([self._html_capture_explanation,
                  self._widget_capture_data,
                  HBox([self._button_recognize_signal_capturefile,
                        self._info_run_botton,
                        self._w_post_process_steps,
                        self._info_validation_method_size],
                       layout=Layout(width='100%')
                       ),
                  self._output_recognition_classification,
                  ]),
            VBox([self._widget_confusion_matrix]),
            VBox([self._widget_feature_summary]),
            VBox([self._widget_model_summary]),



        ])
        self.kb_items = VBox([
            HBox([self._widget_parent_select, self._button_refresh, self._button_rehydrate
                  ]),
            model_explore_tabs,
            self._widget_render_space,
        ])

        model_explore_tabs.set_title(0, 'Test Model')
        model_explore_tabs.set_title(1, 'Confusion Matrix')
        model_explore_tabs.set_title(2, 'Feature Summary')
        model_explore_tabs.set_title(3, 'Model Summary')

        self._output_recognition_classification.layout.visibility = "hidden"

        self._button_refresh.on_click(self._refresh_models_list)
        self._button_rehydrate.on_click(self._rehydrate_model)
        self._widget_parent_select.observe(self._update_confusion_matrix)
        self._button_recognize_signal_capturefile.on_click(
            self._recognize_signal_all)

        self._refresh()

        return self.kb_items


def post_process(results, buffer):

    if buffer > results.shape[0]:
        buffer = results.shape[0]

    post_processed_results = [-1]*(buffer)

    for i in range(0, results.shape[0]-buffer):
        counts = np.bincount(results['Classification'].iloc[i:i+buffer].values)
        post_processed_results.append(np.argmax(counts))

    results['Classification'] = post_processed_results

    return results[results['Classification'] != -1].reset_index(drop=True)



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


from glob import glob
try:
    import configparser as ConfigParser
except:
    import ConfigParser
import os.path


def get_config(config_info_file=None):
    """ Sample Use Case
    ============================================================================
    config_info_file="C:\\Users\\username\\project\\nbi_jv-nbijv\\notebooks\\upload_config.ini"
    data = get_config("your config file")
    files = data[0]   /* file path
    group_names = data[1]  /* ['Accelerator', 'Gyroscrop']
    group_columns = data[2]  /* ['3,4,5', '6,7,8']
    data_column_names = data[3]  /* ['accelx,accely,accelz', 'gyrox,gyroy,gyroz']
    metadata_columns = data[4]  /* [0,1]
    """
    if config_info_file is None:
        raise Exception("*** Sorry, Config File Missing *** " + "file exist? " +
                        str(os.path.isfile(config_info_file)) + '..filename: ' + config_info_file)

    config = ConfigParser.ConfigParser()
    config.read(config_info_file)
    sections = config.sections()
    if "File_Path" in sections[0]:
        filePath = config.get(sections[0], 'path')
    else:
        print("Can not read config file...")

    # First try absolute path
    files = glob(filePath)
    if not files:
        # Then try relative to upload config file
        files = glob(os.path.join(
            os.path.split(config_info_file)[0], filePath))

    for i in range(len(files)):
        if os.path.splitext(files[i])[1] == '.h5':
            return [files, None, None, None, None]

    group_names = config.get(sections[1], 'group_name')
    group_columns = config.get(sections[1], 'group_column')
    data_column_names = config.get(sections[1], 'data_column_name')
    metadata_columns = (config.get(sections[2], 'metadata_column')).split(
        ',') if len(sections) > 2 else None
    groupnames = group_names.split('/')
    groupcolumns = group_columns.split('/')
    datacolumnnames = data_column_names.split('/')
    #metadatacolumns = metadata_columns.split(',')

    return [files, groupnames, groupcolumns, datacolumnnames, metadata_columns]



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017-18  SensiML Corporation.                                   #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


from ipywidgets import HBox, VBox, Dropdown, Button, Image, Layout, Accordion
import numpy as np
from pandas import DataFrame
from bqplot.marks import Scatter, Bars, Lines
from bqplot.scales import LinearScale, OrdinalScale
from bqplot.figure import Figure
from bqplot import Tooltip
from bqplot.axes import Axis
import time
import numpy as np
import threading
import os
from collections import OrderedDict
from qgrid import show_grid
import shutil
import os
import json
from pandas import DataFrame, concat
import copy
from six import text_type



class ModelVisualizationWidget():
    """
    model (knowledpack): a data frame of feature vectors along with a label column and other metadata
    bins (int): the number of bins in the histograms
    label_column (str): the name of the column in the features dataframe that refers to the label infomration
    group_columns (list): if you want other metadata in the tooltip, these columns will be added
    f_lim (dict): this sets the limits for max and min of the plots to a constant
        {'max':10, 'min':10}. otherwise defaults to the values of the current features
        which can be missleading.
    colors (list): list of colors to use. Internally has a list of 10. If the labels
        are longer you will need to pass your own

    """

    def __init__(self, model, feature_vectors=None,  bins=25, label_column=None, group_columns=None, f_lim={'max':255,'min':0}, colors=None, max_vectors=1000):


        if model:
            if hasattr(model, "knowledgepack"):
                model = model.knowledgepack

        self._max_vectors = max_vectors
        self._label_column = label_column
        self._group_columns = group_columns

        if self._label_column is None:
            if hasattr(model, 'query_summary') and model.query_summary:
                self._label_column = model.query_summary['label_column']
            elif hasattr(model, 'pipeline_summary') and model.pipeline_summary:
                self._label_column = model.pipeline_summary[0]['label_column']
            else:
                print("Please specify the label column name")
                return

        if self._group_columns is None:
            if hasattr(model, 'query_summary') and model.query_summary:
                self._group_columns = eval(model.query_summary['metadata_columns'])
            elif hasattr(model, 'pipeline_summary') and model.pipeline_summary:
                self._group_columns = model.pipeline_summary[-1]['ignore_columns']
            else:
                print("Please specify the group column name")
                return

        self._feature_list = [x['Feature'] for x in model.feature_summary]
        self._features = DataFrame(columns=[self._label_column]+self._group_columns+self._feature_list)
        
        if feature_vectors is not None:
            if 'Subject' not in feature_vectors.columns:
                feature_vectors['Subject'] = 1
            self._features = copy.deepcopy(feature_vectors[self._features.columns])
            self._features[self._label_column] = feature_vectors[self._label_column].astype(str)

        self._class_map = OrderedDict({'Unknown':0})

        increment=1
        if '0' in model.class_map:
            increment=0
        for i in range(1, len(model.class_map)+increment):
            self._class_map[str(model.class_map[str(i)])] = i

        self._reverse_class_map = {int(k):str(v) for k,v in model.class_map.items()}
        self._reverse_class_map[0] = 'Unknown'


        if  model.device_configuration['classifier'] == 'PME':
            self._neurons = model.neuron_array
            self._norm = model.device_configuration['distance_mode']
        else:
            self._norm = 0
            self._neurons = []

        self._colors = colors
        self._f_lim = f_lim

        self._bins = bins
        self._counter = 0
        self._hists_y = []
        self._hists_x = []
        self._scatters = []
        self._neuron_scatters = []

        legend_x = []
        legend_y = []
        legend_colors = []

        box_color = 'black'

        if self._colors is None:
            self._colors = ["#16a085","#8e44ad","#d35400","#2c3e50",
                      "#2980b9","#c0392b", "#27ae60", "#6E2C00",
                      "#1A5276", "#17202A"]


        self._feature_x = Dropdown(description='Feature X')
        self._feature_y = Dropdown(description='Feature Y')

        self._feature_x.options = [x for x in self._feature_list]
        self._feature_y.options = [x for x in self._feature_list]

        self._feature_map = {x:index for index,x in enumerate(self._feature_x.options)}

        if len(self._feature_y.options) > 1:
            self._feature_y.value = self._feature_y.options[1]

        feature1 = self._feature_x.value
        feature2 = self._feature_y.value

        if self._f_lim:
            sc_x = LinearScale(min=f_lim['min'], max=f_lim['max'])
            sc_y = LinearScale(min=f_lim['min'], max=f_lim['max'])
        else:
            sc_x = LinearScale()
            sc_y = LinearScale()

        hist_scale_y = LinearScale(min=0)
        hist_scale_x = LinearScale(min=0)
        x_ord_legend = OrdinalScale()
        y_lin_legend = LinearScale()

        feature_tt = Tooltip(fields=['name'], labels=[', '.join(['index', self._label_column]+self._group_columns)])
        neuron_tt = Tooltip(fields=['name'], labels=['Neuron:Category:AIF'])

        self._h_bins = self._get_h_bins()

        for label, index in self._class_map.items():

            legend_x.append(label)
            if self._features.shape[0]:
                legend_y.append((self._features[self._label_column] == label).sum())
            else:
                legend_y.append(1)

            # create a scatter plot for each group
            self._neuron_scatters.append(Scatter(x=[],
                                    y=[],
                                    names=[],
                                    display_names=False,
                                    default_opacities=[0.5],
                                    default_size=30,
                                    scales={'x': sc_x, 'y': sc_y},
                                    colors=[self._colors[index]],
                                    tooltip=neuron_tt,
                                    marker='square'
                                    ))

            # create a scatter plot for each group
            self._scatters.append(Scatter(x=[],
                                    y=[],
                                    names=[],
                                    display_names=False,
                                    default_opacities=[0.5],
                                    default_size=30,
                                    scales={'x': sc_x, 'y': sc_y},
                                    colors=[self._colors[index]],
                                    tooltip=feature_tt,
                                    ))

            # create a histograms using a bar chart for each group
            # histogram plot for bqplot does not have enough options (no setting range, no setting orientation)
            #h_y, h_x = np.histogram(values, bins=h_bins_x)
            self._hists_x.append(Bars(x=[],
                                y=[],
                                opacities=[0.3]*bins,
                                scales={'x': sc_x, 'y': hist_scale_x},
                                colors=[self._colors[index]],
                                orientation='vertical'))

            self._hists_y.append(Bars(x=[],
                                y=[],
                                opacities=[0.3]*bins,
                                scales={'x': sc_x, 'y': hist_scale_y},
                                colors=[self._colors[index]],
                                orientation='horizontal'))

        
        self._neuron_patches = self._init_neurons(feature1, feature2, {"x":sc_x, "y":sc_y})


        # legend will show the names of the labels as well as a total count of each
        self._legend_bar = Bars(x=legend_x,
                   y=legend_y,
                   colors=self._colors,
                   opacities=[0.3]*6,
                   scales={'x': x_ord_legend, 'y':y_lin_legend},
                   orientation='horizontal')

        ax_x_legend = Axis(scale=x_ord_legend,
                    tick_style={'font-size':16},
                    label='',
                    orientation='vertical',
                          )
                    #tick_values=features.groupby(self._label_column).count()[self._feature_list[0]].index)

        ax_y_legend = Axis(scale=y_lin_legend,
                           orientation='horizontal',
                           color='white',
                           num_ticks=0)

        #these are blank blank axis that are used to fill in the boarder for the top and right of the figures
        ax_top = Axis(scale=sc_x, color=box_color, side='top', tick_style={'font-size':0})
        ax_right = Axis(scale=sc_x, color=box_color, side='right', tick_style={'font-size':0})
        ax_left = Axis(scale=sc_x, color=box_color,  side='left', tick_style={'font-size':0})
        ax_bottom = Axis(scale=sc_x, color=box_color, side='bottom', tick_style={'font-size':0})

        #scatter plot axis
        self._ax_x = Axis(label=feature1, scale=sc_x, color=box_color)
        self._ax_y = Axis(label=feature2, scale=sc_y, orientation='vertical', color=box_color)

        #count column of histogram
        ax_count_vert  = Axis(label='', scale=hist_scale_y, orientation='vertical', color=box_color, num_ticks=3)
        ax_count_horiz = Axis(label='', scale=hist_scale_x, orientation='horizontal', color=box_color, num_ticks=3)

        #histogram bin axis
        ax_hist_x = Axis(label='', scale=sc_x, orientation='vertical', color=box_color)
        ax_hist_y = Axis(label='', scale=sc_x, orientation='horizontal', color=box_color)

        #create figures for each plot
        f_scatter = Figure(axes=[self._ax_x, self._ax_y, ax_top, ax_right],
                           background_style={'fill':'white'}, #css is inserted directly
                           marks=self._neuron_patches+self._neuron_scatters+self._scatters,
                           min_aspect_ratio=1,
                           max_aspect_ratio=1,
                           fig_margin =  {"top":0, "bottom":60, "left":60, "right":0},
                           )

        f_hists_y = Figure(axes=[ax_left, ax_count_horiz, ax_top, ax_right],
                           background_style={'fill':'white'},
                           marks=self._hists_y,
                           min_aspect_ratio=.33,
                           max_aspect_ratio=.33,
                           fig_margin =  {"top":0, "bottom":60, "left":10, "right":0},
                          )

        f_hists_x = Figure(axes=[ax_count_vert, ax_top, ax_right],
                           background_style={'fill':'white'},
                           marks=self._hists_x,
                           min_aspect_ratio=3,
                           max_aspect_ratio=3,
                           fig_margin =  {"top":20, "bottom":10, "left":60, "right":0},
                          )

        f_legend = Figure(marks=[self._legend_bar], axes=[ax_x_legend, ax_y_legend], title='',
                                     legend_location ='bottom-right',
                                     background_style = {'fill':'white'},
                                     #min_aspect_ratio=1,
                                     #max_aspect_ratio=1,
                                     fig_margin =  {"top":20, "bottom":30, "left":120, "right":20})

        try:                               
            dir_path = os.path.dirname(os.path.realpath(__file__))

            file = open(os.path.join(dir_path,'..','image',"sensiml.png"), "rb")
            image = file.read()
            sensiml_logo = Image(
                value=image,
                format='png',
                width=300,
                height=400,
            )
        except:
            sensiml_logo=Image(width=300,height=400)

        self._grid = show_grid(self._features, grid_options={'maxVisibleRows':10, 'minVisibleRows':5})
        # we already set the ratios, but it is necessary to set the size explicitly anyway
        # this is kind of cool, inserts this into the style in html
        
        dist = text_type("480px")
        sec_dist = text_type("240px")
        third_dist = text_type("160px")

        f_legend.layout.height =  text_type("{}px".format(50*len(legend_y)))
        f_legend.layout.width = sec_dist
        f_hists_x.layout.height = third_dist
        f_hists_x.layout.width = dist
        f_hists_y.layout.height = dist
        f_hists_y.layout.width = third_dist
        f_scatter.layout.height = dist
        f_scatter.layout.width = dist

        accordion= Accordion(children=[self._grid])
        accordion.set_title(0, 'Feature Vectors')
        accordion.selected_index = None

        # when the user selects a different feature, switch the data plotted
        self._feature_x.observe(self._change_x_feature, 'value')
        self._feature_y.observe(self._change_y_feature, 'value')

        self._change_x_feature(None)
        self._change_y_feature(None)

        #display the widgets'
        self._widget = VBox([
                    HBox([
                        VBox([sensiml_logo,
                               f_legend,
                               self._feature_x,
                               self._feature_y,
                              ]),
                        VBox([
                           HBox([f_hists_x]),
                          HBox([f_scatter, f_hists_y])
                            ]),

                        ]),

                    accordion],
                        layout = Layout(border='outset',
                width='100%'))

        display(self._widget)

    # we create some functions that allow changes when the widgets notice an event
    def _change_x_feature(self, b):
        feature_groups = self._features.groupby([self._label_column])
        for label, index in self._class_map.items():
            if label not in feature_groups.groups.keys():
                continue
            group = feature_groups.get_group(label)
            self._scatters[index].y = group[self._feature_y.value]
            self._scatters[index].x = group[self._feature_x.value]
            try:
                if not self._scatters[index].names:
                    self._scatters[index].names = self._get_names(group)
            except:
                pass


            h_y, h_x = np.histogram(group[self._feature_x.value].values, bins=self._h_bins)
            self._hists_x[index].y = h_y
            self._hists_x[index].x = h_x+self._hist_shift

        self._update_neurons(self._feature_x.value, self._feature_y.value)

        self._ax_x.label = self._feature_x.value

    def _change_y_feature(self, b):
        feature_groups = self._features.groupby([self._label_column])
        for label, index in self._class_map.items():
            if label not in feature_groups.groups.keys():
                continue
            group = feature_groups.get_group(label)
            self._scatters[index].y = group[self._feature_y.value]
            self._scatters[index].x = group[self._feature_x.value]
            h_y, h_x = np.histogram(group[self._feature_y.value].values, bins=self._h_bins)
            self._hists_y[index].y = h_y
            self._hists_y[index].x = h_x+self._hist_shift


        self._update_neurons(self._feature_x.value, self._feature_y.value)

        self._ax_y.label = self._feature_y.value

    def _get_names(self, f):
        M = []
        for i in range(len(f)):
            self._counter+=1
            M.append('{},'.format(self._counter)+',   '.join([str(x) for x in f[[self._label_column]+self._group_columns].iloc[i].values]))
        return M


    def _init_neurons(self, feature_x, feature_y, scale):
        neuron_patches = []
        self._init_neuron_names()

        for neuron in self._neurons:
            x = neuron['Vector'][self._feature_map[feature_x]]
            y = neuron['Vector'][self._feature_map[feature_y]]
            aif   = neuron['AIF']

            if neuron['Category'] < 30000: # if it is not degenerated neuron
                clr = self._colors[neuron['Category']]
                if self._norm == 0:
                    dim = len(neuron['Vector'])
                    aif = 2*aif/dim
                    neuron_patches.append(Lines( x=[x,x+aif,x,x-aif],
                                                y=[y-aif,y,y+aif,y],
                                                opacities=[.2],
                                                fill='inside',
                                                colors=[clr],
                                                scales=scale))
                else:
                    aif=aif/2
                    neuron_patches.append(Lines(x=[x-aif,x+aif,x+aif,x-aif],
                                                y=[y-aif,y-aif,y+aif,y+aif],
                                                opacities=[.2],
                                                fill='inside',
                                                colors=[clr],
                                                scales=scale))

        return neuron_patches

    def _update_neurons(self, feature_x, feature_y):
        if not self._neurons:
            return 

        for index, neuron in enumerate(self._neurons):
            x = neuron['Vector'][self._feature_map[feature_x]]
            y = neuron['Vector'][self._feature_map[feature_y]]
            aif = neuron['AIF']

            if neuron['Category'] < 30000: # if it is not degenerated neuron
                if self._norm == 0:
                    dim = len(neuron['Vector'])
                    aif = 2*aif/dim
                    self._neuron_patches[index].x = [x,x+aif,x,x-aif]
                    self._neuron_patches[index].y =[y-aif,y,y+aif,y]
                else:
                    aif=aif/2
                    self._neuron_patches[index].x=[x-aif,x+aif,x+aif,x-aif]
                    self._neuron_patches[index].y=[y-aif,y-aif,y+aif,y+aif]

        neuron_groups = self._neuron_df.groupby([self._label_column])

        for label, index in self._class_map.items():
            if label in neuron_groups.groups.keys():
                self._neuron_scatters[index].y = neuron_groups.get_group(label)[feature_y].values
                self._neuron_scatters[index].x = neuron_groups.get_group(label)[feature_x].values

    # simple function to return the bins for the plot
    def _get_h_bins(self):
        self._hist_shift = 0
        if self._f_lim:
            delta = (self._f_lim['max']-self._f_lim['min'])/float(self._bins)
            self._hist_shift = delta/2
            return np.arange(self._f_lim['min'], self._f_lim['max']+delta,delta)

    def _init_neuron_names(self):
        if not self._neurons:
            return

        neuron_centers = []
        names = [[] for i in range(len(self._class_map.keys()))]
        for index, neuron in enumerate(self._neurons):
            neuron_center = {self._feature_list[i]:neuron['Vector'][i] for i in range(len(self._feature_list))}
            neuron_center[self._label_column] = self._reverse_class_map[neuron['Category']]
            neuron_centers.append(neuron_center)
            names[neuron['Category']].append('{0}:{1}:{2}'.format(neuron['Identifier'],neuron['Category'],neuron['AIF']))

        self._neuron_df = DataFrame(neuron_centers)
        
        neuron_groups = self._neuron_df.groupby([self._label_column])

        for label, index in self._class_map.items():
            if label in neuron_groups.groups.keys():
                self._neuron_scatters[index].names = names[index]



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


from warnings import warn


class PipelineWarning(Warning):
    """Base class for a pipeline warning"""
    pass


class PipelineError(Exception):
    """Base class for a pipeline error"""
    pass


class PipelineStep(object):
    """Base class for a pipeline step"""

    def __init__(self, name, step_type='default'):
        self._inputs = []
        self._outputs = []
        self._name = name
        self._step_type = step_type
        self._feature_table = None

    @property
    def feature_table(self):
        return self._feature_table

    @feature_table.setter
    def feature_table(self, value):
        self._feature_table = value

    @property
    def inputs(self):
        return self._inputs

    @property
    def outputs(self):
        return self._outputs

    @outputs.setter
    def outputs(self, outputs):
        self._outputs = outputs

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def step_type(self):
        return self._step_type

    def _check_inputs(self, output_vars=()):
        """Compares the list of inputs with the given output_vars list.

        Args:
            output_vars (list[str]): strings representing the current set of output variables

        Returns:
            (list): unmatched inputs, or an empty list if all inputs are matched
        """
        unmatched_inputs = []
        for input in self.inputs:
            if not input in output_vars:
                unmatched_inputs.append(input)
        return unmatched_inputs

    def _check_outputs(self, output_vars=(), is_last=False):
        """Compares the outputs from the item with the other output_vars from the pipeline.

        Args:
            output_vars (list[str]): strings representing the current set of output variables
            is_last (bool): True if item is the last step, otherwise False (default)

        Returns:
            (list): duplicate outputs, or an empty list if all outputs are unique

        Raises:
            PipelineError if step is not the last one and does not have any outputs
            PipelineWarning if step is the last one and does not have any outputs
        """
        duplicate_outputs = []
        if self.outputs:
            for output in self.outputs:
                if output in output_vars:
                    duplicate_outputs.append(output)
        elif is_last:
            warn('{0} {1} does not have any outputs.'.format(
                self._step_type, self._name), PipelineWarning)
        else:  # is_last = False
            raise PipelineError('{0} {1} does not have any outputs.'.format(
                self._step_type, self._name))
        return duplicate_outputs



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017-18  SensiML Corporation.                                   #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


import os
import zipfile
from os import listdir, makedirs
from os.path import isfile, join, exists, splitext
from pathlib import Path

from ipywidgets import Layout, VBox, HBox, Button
from ipywidgets import widgets

from sensiml.device_flashing import *
from sensiml.widgets.base_widget import BaseWidget


class FlashWidget(BaseWidget):
    def __init__(self, dsk=None, folder='knowledgepacks'):
        self._dsk = dsk
        self._folder = Path(folder)
        self._flashing_file = ''
        self._selected_flasher = None
        if not exists(folder):
            makedirs(folder)

    def _refresh_platform_files(self, b=None):
        key_name = self.platform_widget.value.replace(' ', '-')
        platform_search = '{}'.format(key_name).lower()
        ext_match = '.zip'
        self.flashable_files_widget.options = []
        platform_files = [f for f in listdir(self._folder) if isfile(join(self._folder, f))
                          and platform_search in f.lower()
                          and splitext(f)[-1] == ext_match
                          and 'bin' in f.lower()]

        if len(platform_files) > 0:
            self.flashable_files_widget.options = platform_files
            self.flashable_files_widget.value = platform_files[0]

    def _refresh_flash_methods(self):
        if self._selected_flasher is None:
            return
        self.flash_methods_widget.options = self._selected_flasher.flash_methods

    def select_platform(self, platform_name):
        if isinstance(platform_name, int):
            self._selected_flasher = self._flashers[platform_name]
        elif isinstance(platform_name, str):
            for f in self._flashers:
                if platform_name in f.supported_platforms:
                    self._selected_flasher = f
                    break
        else:
            print("Unknown Platform: {}".format(str(platform_name)))
            return
        self._refresh_platform_files()
        self._refresh_flash_methods()

    def get_platform_names(self):
        flasher_boards = list()
        for j in self._flashers:
            flasher_boards.extend(j.supported_platforms)
        return flasher_boards

    def on_flash_button_clicked(self, b):
        if not self.flashable_files_widget.value:
            return
        if not self._selected_flasher:
            return
        flash_zip_file = Path(self._folder, self.flashable_files_widget.value)
        self._flashing_file = unzip_folder(flash_zip_file, self._selected_flasher.flash_file_ext)

        self._selected_flasher.flash(self._flashing_file, method=self.flash_methods_widget.value)

    def _refresh(self, b=None):
        if self._dsk is None:
            return
        self.platform_widget.options = self.get_platform_names()

    def create_widget(self):

        self._flashers = [NordicThingyFlasher(), STDeviceFlasher()]

        self.platform_widget = widgets.Dropdown(description='Platform')
        self.description_widget = widgets.Label(layout=Layout(width='66%'))
        self.flashable_files_widget = widgets.Dropdown(
            description='Binary', layout=Layout(width='85%'))
        self.flash_methods_widget = widgets.Dropdown(
            description="Flash Method")
        self.flash_button = Button(description="Flash")
        self.refresh_button = Button(
            icon='refresh', tooltip='Refresh File List', layout=Layout(width='15%'))

        self._refresh()

        self.flash_button.on_click(self.on_flash_button_clicked)
        self.refresh_button.on_click(self._refresh_platform_files)

        return VBox([
            HBox([widgets.interactive(self.select_platform,
                                      platform_name=self.platform_widget), self.description_widget]),
            HBox([self.flashable_files_widget, self.refresh_button]),
            HBox([self.flash_methods_widget, self.flash_button])
        ])


def unzip_folder(path_to_zip_file, file_ext):
    path_string = str(path_to_zip_file.absolute())
    base_name = os.path.basename(path_string[:-4])
    binary_folder = Path(path_string[:-4])
    binary_path = Path(path_string[:-4], base_name + file_ext)

    zip_ref = zipfile.ZipFile(path_to_zip_file, 'r')
    zip_ref.extractall(binary_folder)
    zip_ref.close()

    return binary_path



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################

import warnings


class DeviceConfig(dict):
    """
    Defines dictionary class for device configuration to be sent from sensiml to the
    Server for KnowledgePack code/binary generation.
    """

    # List of fields and their defaults
    fields = {
        'target_platform': 0,
        'build_flags': '',
        'budget': {},
        'debug': False,
        'sram_size': None,
        'test_data': '',
        'application': '',
        'sample_rate': 100,
        'kb_description': None,
    }

    def __init__(self, iterable=None, strict=True, **kwargs):
        super(DeviceConfig, self).__init__()
        object.__setattr__(self, '_strict', strict)
        for k, v in self.fields.items():
            self[k] = v
        if iterable:
            for k, v in dict(iterable).items():
                setattr(self, k, v)
        if kwargs:
            for k, v in kwargs.items():
                setattr(self, k, v)

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        if self._strict and name not in self.fields.keys():
            out_str = '{} not in device configuration.'.format(name)
            out_str += '\nYou may need to update the device configuration on KB Cloud.'
            warnings.warn(out_str, DeprecationWarning)

            return
        if name == 'budget':
            assert isinstance(value, dict), 'Budget is not a dictionary.'
        elif name == 'debug':
            value = bool(value)

        self[name] = value

    def initialize_from_dict(self, init_dict):
        """DEPRECATED: Populates a single device config from a dictionary of properties.

            Args:
                (dict): contains target_platform, platform_version, build_flags, budget, debug and sram_size
        """
        warnings.warn(
            'initialize_from_dict() is deprecated. Please pass arguments to constructor', DeprecationWarning)
        for k, v in init_dict.items():
            setattr(self, k, v)



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################

import os
import json
from pandas import DataFrame, Series

from sensiml.base import utility
from sensiml.datamanager.sandbox import describe_pipeline


class PipelineSeed(object):
    """Base class for an automation pipeline seed object"""
    _id = 0
    _name = ''
    _description = ''
    _pipeline = ''
    _input_contract = {}

    def __init__(self, connection):
        self._connection = connection

    @property
    def name(self):
        return self._name

    @property
    def id(self):
        return self._id

    @property
    def description(self):
        return self._description

    @property
    def pipeline(self):
        return self._pipeline

    @property
    def input_contract(self):
        return self._input_contract

    def refresh(self):
        """Calls the REST API and populates the local object properties from the server."""
        url = 'seed/{0}/'.format(self.id)
        response = self._connection.request('get', url,)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self._name = response_data['name']
            self._description = response_data['description']
            self._id = response_data['id']
            self._pipeline = response_data['pipeline']
            self._input_contract = response_data['input_contract']

    def initialize_from_dict(self, input_dictionary):
        """Populates a single seed object from a dictionary of properties from the server."""
        self._id = input_dictionary['id']
        self._name = input_dictionary['name']
        self._description = input_dictionary['description']
        self._pipeline = input_dictionary['pipeline']
        self._input_contract = input_dictionary['input_contract']

    def __dict__(self):
        ret = {
            'Name': self.name,
            'Pipeline': self.pipeline,
            'Description': self.description,
            'Inputs': self.input_contract,
        }
        return ret

    def __call__(self):
        pd_dict = self.__dict__()
        pseries = Series(pd_dict, index=pd_dict.keys())
        df = DataFrame()
        df = df.append(pseries, ignore_index=True)
        return df.set_index('Name').reset_index()

    def describe(self, show_params=True):
        """Prints a formatted description of the pipeline seed"""
        describe_pipeline(self.pipeline, show_params=show_params)



import subprocess

class DeviceFlasher(object):
    def __init__(self):
        self._flash_methods = list()
        self._name = ''
        self._supported_platforms = list()
        self._needs_scan = False
        self._description=''
        self._flash_file_ext = '.bin'

    @property
    def name(self):
        return self._name

    @property
    def flash_methods(self):
        return self._flash_methods

    @property
    def supported_platforms(self):
        return self._supported_platforms

    @property
    def needs_scan(self):
        return self._needs_scan

    @property
    def description(self):
        return self._description

    @property
    def flash_file_ext(self):
        return self._flash_file_ext

    def _wait_for_process(self, command):
        print('Running the following command: ')
        print(command)
        sub_proc = subprocess.Popen(command[0], stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        p_val = sub_proc.poll()
        while (p_val is None):
            print(sub_proc.stdout.read().decode("UTF-8"))
            print(sub_proc.stderr.read().decode("UTF-8"))
            p_val = sub_proc.poll()

        if p_val is 0:
            print("Process Completed successfully.")
        else:
            print("And error occured. Error code {}".format(p_val))


##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017-18  SensiML Corporation.                                   #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


class DskException(Exception):
    """Base KB_DSK_Basic Exception"""
    pass


class QueryExistsException(DskException):
    pass


class BadArgumentException(DskException):
    pass


class BackendErrorException(DskException):
    pass


class FeatureFileExistsException(DskException):
    def __init__(self):
        Exception.__init__(
            self, 'FeatureFile already exists. Use force=True to override.')


class PipelineOrderException(Exception):
    pass


class PipelineDataColumnsException(Exception):
    def __init__(self):
        msg = """Warning: Binary generation is currently only possible for the following sensors:
* ACCELEROMETERX
* ACCELEROMETERY
* ACCELEROMETERZ
* GYROSCOPEX
* GYROSCOPEY
* GYROSCOPEZ
If you are using these sensors but have different names, change the column
names of your data to align with the above to generate device code."""
        print(msg)


class DuplicateValueError(Exception):
    def __init__(self, message='Duplicate values used'):
        super(DuplicateValueError, self).__init__(message)



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


import json
from pandas import DataFrame, Series

from sensiml.datamanager.pipelineseed import PipelineSeed
import sensiml.base.utility as utility


class PipelineSeeds():
    """Base class for a collection of pipeline seeds"""

    def __init__(self, connection):
        self._connection = connection
        self.build_pipeline_seeds()

    def __getitem__(self, index):
        return self.seed_list[index]

    def build_pipeline_seeds(self):
        """Populates the seed_list property from the server."""
        self.seed_list = []
        self.seed_dict = {}

        seeds = self.get_seeds()
        for seed in seeds:
            self.seed_dict[seed.name] = seed
            self.seed_list.append(seed)

    def get_seed_by_name(self, name):
        if len(self.seed_list) == 0:
            self.build_pipeline_seeds()
        return self.seed_dict.get(name, None)

    def _new_seed_from_dict(self, dict):
        """Creates and populates a seed from a set of properties.

            Args:
                dict (dict): contains properties of a Seed

            Returns:
                pipeline seed
        """
        seed = PipelineSeed(self._connection)
        seed.initialize_from_dict(dict)
        return seed

    def get_seeds(self):
        """Gets all seeds as seed objects.

            Returns:
                list of functions
        """
        url = 'seed/v2/'
        response = self._connection.request('get', url,)
        try:
            response_data, err = utility.check_server_response(response)
        except ValueError:
            print(response)

        pipeline_seeds = []
        for seed in response_data:
            pipeline_seeds.append(self._new_seed_from_dict(seed))

        return pipeline_seeds

    def __call__(self):
        all_seeds = self.get_seeds()
        if (len(all_seeds) < 0):
            return DataFrame()
        ret = DataFrame(columns=['Name', 'Description'])
        for seed in all_seeds:
            ret = ret.append(seed(), ignore_index=True)
        ret = ret.set_index('Name').reset_index()
        return ret

    def __str__(self):
        output_string = 'Seeds are pipeline templates on the server that can be used to generate models \n' + \
                        'using the dsk.pipeline.auto() command. They eliminate the need to specify a \n' + \
                        'pipeline and are most effective when chosen appropriately for the application.\n\n'

        for seed in sorted(self.seed_list, key=lambda x: x.id):
            output_string += 'Name:          {0}\n'.format(seed.name)
            pipeline_format = '{} -> ' * len(seed.pipeline)
            output_string += 'Pipeline:      {0}\n\n'.format(
                pipeline_format.format(*[step['name'] for step in seed.pipeline])[:-3])
            output_string += '{0}\n\n'.format(seed.description)

        output_string += '\n'

        return output_string
